#!/usr/bin/python3

import asyncio

from camisole.utils import indent, tabulate


async def print_working_languages(languages, verbosity):
    from camisole.ref import test
    from pprint import pformat
    import sys

    def extract_fail_message(result):
        relevant = (result['compile'] if 'compile' in result
                    else result['tests'][0])
        messages = [relevant['meta']['message'], relevant['stderr']]
        return '\n'.join(msg for msg in messages if msg)

    use_color = sys.stdout.isatty()
    max_length = max(map(len, languages)) + 2

    res = [asyncio.Task(test(lang_name)) for lang_name in languages]
    await asyncio.wait(res)

    for i, lang_name in enumerate(languages):
        success, raw = res[i].result()
        status, color = ('OK', 32) if success else ('FAIL', 31)
        ok_msg = (f'\x1B[{color}m{status}\033[0m' if use_color else status)
        print(f'{lang_name + " ":.<{max_length}} {ok_msg}', flush=True)

        if not success and verbosity > 0:
            if verbosity == 1:
                print(indent(extract_fail_message(raw).strip()))
            else:
                print(indent(pformat(raw)))


def main():
    import argparse
    import asyncio
    import camisole.loader
    import camisole.http
    import logging
    import os

    parser = argparse.ArgumentParser(
        description="asyncio-based source compiler and test runner")
    parser.add_argument(
        '-l',
        '--logging',
        default='error',
        choices=[l.lower() for l in logging._nameToLevel],
        help="logging level")
    parser.add_argument(
        '-m',
        '--module',
        dest='modules',
        action='append',
        help="extra modules to load (customize search path with CAMISOLEPATH)")

    cmd = parser.add_subparsers(dest='command')

    parse_serve = cmd.add_parser('serve', add_help=False)
    parse_serve.add_argument('-h', '--host', default='0.0.0.0')
    parse_serve.add_argument('-p', '--port', type=int, default=8080)
    parse_serve.add_argument('--help', action='help')

    parse_languages = cmd.add_parser('languages')

    parse_test = cmd.add_parser('test')
    parse_test.add_argument(
        '-v',
        dest='verbose',
        action='append_const',
        const=1,
        help="increase verbosity (up to two)")
    parse_test.add_argument(
        'languages',
        nargs=argparse.REMAINDER,
        help="languages to test")
    args = parser.parse_args()

    logging.basicConfig(level=args.logging.upper())

    # import built-in languages
    camisole.languages._import_builtins()

    # import extra languages
    if args.modules:
        path = os.environ.get('CAMISOLEPATH', '')
        camisole.loader.load_modules(args.modules, path)

    loop = asyncio.get_event_loop()

    if args.command == 'languages':
        from camisole.languages import all
        rows = [(lang, cls.name, cls.__module__, cls.__name__)
                for lang, cls in sorted(all().items())]
        print("\n".join(tabulate(rows,
            headers=("Name", "Display name", "Module", "Class name"))))

    elif args.command == 'test':
        from camisole.languages import all
        languages = args.languages or all().keys()
        verbosity = sum(args.verbose or [])
        loop.run_until_complete(print_working_languages(languages, verbosity))

    elif args.command == 'serve':
        from camisole.languages import all
        logging.info("Registry has %d languages:\n%s", len(all()),
                     '\n'.join(f'    {l!r}' for l in all().values()))
        camisole.http.run(host=args.host, port=args.port)

    else:
        parser.exit(message=parser.format_usage())


if __name__ == '__main__':
    main()



import importlib
import logging
import sys
from itertools import chain
from pathlib import Path
from typing import Iterable

from camisole.languages import all
from camisole.utils import uniquify

logger = logging.getLogger(__name__)

DEFAULT_SEARCH_PATH = [
    Path('~/.local/share/camisole/lang').expanduser(),
    Path('/usr/share/camisole/lang'),
]


def parse_search_paths(paths: str) -> Iterable[Path]:
    paths = (Path(part).resolve() for part in paths.split(':')
             if part)
    return uniquify(chain(paths, DEFAULT_SEARCH_PATH))


def load_modules(modules: Iterable[str], extra_paths: str = ''):
    extra_paths = parse_search_paths(extra_paths)
    if extra_paths:
        sys.path[0:0] = map(str, extra_paths)

    found = set()
    for name in uniquify(modules):
        langs_before = set(all().values())
        try:
            module = importlib.import_module(name)
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                f"Could not find module '{name}' in path") from None
        diff = set(all().values()) - langs_before
        found |= diff
        if diff:
            logger.debug("Module '%s' provides %s", module.__name__,
                         ", ".join(map(repr, diff)))
    return found



from camisole.models import Lang, Program


class FSharp(Lang, name="F#"):
    source_ext = '.fs'
    compiler = Program('fsharpc', opts=['-O'], version_opt='--help')
    interpreter = Program('mono')
    reference_source = r'''
#light
open System
let () =
    Printf.printf "42\n"
'''



from camisole.models import Lang, Program


class Pascal(Lang):
    source_ext = '.pas'
    compiler = Program('fpc', opts=['-XD', '-Fainitc'],
                       version_opt='-h', version_lines=1)
    reference_source = r'''
program main;
begin
    Writeln(42);
end.
'''

    def compile_opt_out(self, output):
        return ['-o' + output]



from camisole.models import Lang, Program


class C(Lang):
    source_ext = '.c'
    compiler = Program('gcc',
                       opts=['-std=c11', '-Wall', '-Wextra', '-O2', '-lm'])
    reference_source = r'''
#include <stdio.h>

int main(void)
{
    printf("42\n");
    return 0;
}
'''



# This file is part of Camisole.
#
# Copyright (c) 2016 Antoine Pietri <antoine.pietri@prologin.org>
# Copyright (c) 2016 Alexandre Macabies <alexandre.macabies@prologin.org>
# Copyright (c) 2016 Association Prologin <info@prologin.org>
#
# Camisole is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Prologin-SADM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Prologin-SADM.  If not, see <http://www.gnu.org/licenses/>.

import functools
import logging
import os
import re
import subprocess
import tempfile
import warnings
from pathlib import Path

import camisole.isolate
import camisole.utils


class Program:
    def __init__(self, cmd, *, opts=None, env=None,
                 version_opt='--version', version_lines=1,
                 version_regex='\d+(\.\d+)+'):
        self.cmd = camisole.utils.which(cmd)
        self.cmd_name = cmd
        self.opts = opts or []
        self.env = env or {}
        self.version_opt = version_opt
        self.version_lines = version_lines
        self.version_regex = re.compile(version_regex)

    @functools.lru_cache()
    def _version(self):
        if self.version_opt is None:
            return None
        proc = subprocess.run([self.cmd, self.version_opt],
                              stderr=subprocess.STDOUT, stdout=subprocess.PIPE)
        return proc.stdout.decode().strip()

    def version(self):
        if self.version_opt is None:
            return None
        res = self.version_regex.search(self._version())
        return res[0] if res else None

    def long_version(self):
        if self.version_opt is None:
            return None
        return '\n'.join(self._version().split('\n')[:self.version_lines])


class MetaLang(type):
    """Metaclass to customize Lang subclasses __repr__()"""

    def __repr__(self):
        return "<{realname}{name}>".format(
            realname=self.__name__,
            name=f' “{self.name}”' if self.__name__ != self.name else '')


class Lang(metaclass=MetaLang):
    """
    Abstract language descriptor.

    Subclass and define the relevant attributes and methods, if need be.
    """
    _registry = {}

    source_ext = None
    compiler = None
    interpreter = None
    allowed_dirs = []
    extra_binaries = {}
    reference_source = None

    def __init_subclass__(cls, register=True, name=None, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.name = name or cls.__name__

        if not register:
            return

        for binary in cls.required_binaries():
            if binary is not None and not os.access(binary.cmd, os.X_OK):
                logging.info(f'{cls.name}: cannot access `{binary.cmd}`, '
                             'language not loaded')
                return

        registry_name = cls.name.lower()
        if registry_name in cls._registry:
            full_name = lambda c: f"{c.__module__}.{c.__qualname__}"
            warnings.warn(f"Lang registry: name '{registry_name}' for "
                          f"{full_name(cls)} overwrites "
                          f"{full_name(Lang._registry[registry_name])}")
        cls._registry[registry_name] = cls

    def __init__(self, opts):
        self.opts = opts

    @classmethod
    def required_binaries(cls):
        if cls.compiler:
            yield cls.compiler
        if cls.interpreter:
            yield cls.interpreter
        yield from cls.extra_binaries.values()

    @classmethod
    def programs(cls):
        return {p.cmd_name: {'version': p.version(), 'opts': p.opts}
                for p in cls.required_binaries()}

    async def compile(self):
        if not self.compiler:
            raise RuntimeError("no compiler")

        # We give compilers a nice /tmp playground
        root_tmp = tempfile.TemporaryDirectory(prefix='camisole-tmp-')
        os.chmod(root_tmp.name, 0o777)
        tmparg = [f'/tmp={root_tmp.name}:rw']

        isolator = camisole.isolate.Isolator(
            self.opts.get('compile', {}),
            allowed_dirs=self.allowed_dirs + tmparg)
        async with isolator:
            wd = Path(isolator.path)
            source = wd / self.source_filename()
            compiled = wd / self.execute_filename()
            with source.open('w') as sourcefile:
                sourcefile.write(self.opts.get('source', ''))
            cmd = self.compile_command(str(source), str(compiled))
            await isolator.run(cmd, env=self.compiler.env)
            binary = self.read_compiled(str(compiled), isolator)

        root_tmp.cleanup()

        return (isolator.isolate_retcode, isolator.info, binary)

    async def execute(self, binary, opts=None):
        if opts is None:
            opts = {}
        opts = {**self.opts.get('execute', {}), **opts}
        input_data = None
        if 'stdin' in opts and opts['stdin']:
            input_data = opts['stdin'].encode()

        isolator = camisole.isolate.Isolator(opts,
                                             allowed_dirs=self.allowed_dirs)
        async with isolator:
            wd = isolator.path
            compiled = self.write_binary(Path(wd), binary)
            env = self.interpreter.env if self.interpreter else None
            await isolator.run(self.execute_command(str(compiled)),
                               env=env, data=input_data)
        return (isolator.isolate_retcode, isolator.info)

    async def run_compilation(self, result):
        if self.compiler is not None:
            cretcode, info, binary = await self.compile()
            result['compile'] = info
            if cretcode != 0:
                return
            if binary is None:
                if result['compile']['stderr'].strip():
                    result['compile']['stderr'] += '\n\n'
                result['compile']['stderr'] += 'Cannot find result binary.\n'
                return
        else:
            binary = self.opts.get('source', '').encode()
        return binary

    async def run_tests(self, binary, result):
        tests = self.opts.get('tests', [{}])
        if tests:
            result['tests'] = [{}] * len(tests)
        for i, test in enumerate(tests):
            retcode, info = await self.execute(binary, test)
            result['tests'][i] = {
                'name': test.get('name', 'test{:03d}'.format(i)),
                **info
            }

            if retcode != 0 and (
                    test.get('fatal', False) or
                    self.opts.get('all_fatal', False)):
                break

    async def run(self):
        result = {}
        binary = await self.run_compilation(result)
        if not binary:
            return result
        await self.run_tests(binary, result)
        return result

    def compile_opt_out(self, output):
        return ['-o', output]

    def read_compiled(self, path, isolator):
        try:
            with Path(path).open('rb') as c:
                return c.read()
        except (FileNotFoundError, PermissionError):
            pass

    def write_binary(self, path, binary):
        compiled = path / self.execute_filename()
        with compiled.open('wb') as c:
            c.write(binary)
        compiled.chmod(0o700)
        return compiled

    def source_filename(self):
        return 'source' + self.source_ext

    def execute_filename(self):
        return 'compiled'

    @staticmethod
    def filter_box_prefix(s):
        return re.sub('/var/(local/)?lib/isolate/[0-9]+', '', s)

    def compile_command(self, source, output):
        if self.compiler is None:
            return None
        return [self.compiler.cmd,
                *self.compiler.opts,
                *self.compile_opt_out(self.filter_box_prefix(output)),
                self.filter_box_prefix(source)]

    def execute_command(self, output):
        cmd = []
        if self.interpreter is not None:
            cmd += [self.interpreter.cmd] + self.interpreter.opts
        return cmd + [self.filter_box_prefix(output)]


class PipelineLang(Lang, register=False):
    """
    A meta-language that compiles multiple sub-languages, passing the
    compilation result to the next sub-language, and eventually executing the
    last result.

    Subclass and define the ``sub_langs`` attribute.
    """
    sub_langs = []

    @classmethod
    def required_binaries(cls):
        yield from super().required_binaries()
        for sub in cls.sub_langs:
            yield from sub.required_binaries()

    async def run_compilation(self, result):
        source = self.opts.get('source', '')
        for i, lang_cls in enumerate(self.sub_langs):
            lang = lang_cls({**self.opts, 'source': source})
            cretcode, info, binary = await lang.compile()
            result['compile'] = info
            if cretcode != 0:
                return
            if binary is None:
                if result['compile']['stderr'].strip():
                    result['compile']['stderr'] += '\n\n'
                result['compile']['stderr'] += 'Cannot find result binary.\n'
                return
            # compile output is next stage input
            source = binary
            if i < len(self.sub_langs) - 1:
                source = source.decode(errors='ignore')
        return binary

    async def compile(self):
        raise NotImplementedError()



from camisole.models import Lang, Program


class Haskell(Lang):
    source_ext = '.hs'
    compiler = Program('ghc', opts=['-dynamic', '-O2'])
    reference_source = r'module Main where main = putStrLn "42"'



import importlib
from typing import Mapping, Type

from camisole.models import Lang


def all() -> Mapping[str, Type[Lang]]:
    return Lang._registry


def by_name(name: str) -> Type[Lang]:
    return all()[name.lower()]


def _import_builtins():
    for name in __all__:
        importlib.import_module(f'{__name__}.{name}')


__all__ = [
    'ada',
    'brainfuck',
    'c',
    'csharp',
    'cxx',
    'fsharp',
    'haskell',
    'java',
    'javascript',
    'lua',
    'ocaml',
    'pascal',
    'perl',
    'php',
    'python',
    'ruby',
    'rust',
    'scheme',
    'visualbasic',
]



from camisole.models import Lang, Program


class CSharp(Lang, name="C#"):
    source_ext = '.cs'
    compiler = Program('mcs', opts=['-optimize+'])
    interpreter = Program('mono')
    reference_source = r'''
using System;
class Program
{
    public static void Main()
    {
        Console.WriteLine(42);
    }
}
'''

    def compile_opt_out(self, output):
        return ['-out:' + output]



import jsonschema


ISOLATE_OPTS_PROPERTIES = {
    'fsize': {'type': 'number'},
    'mem': {'type': 'number'},
    'processes': {'type': 'number'},
    'quota': {'type': 'string'},
    'stack': {'type': 'string'},
    'time': {'type': 'number'},
    'wall-time': {'type': 'number'},
}

EXECUTE_PROPERTIES = {
    'stdin': {'type': 'string'},
    **ISOLATE_OPTS_PROPERTIES,
}

SCHEMA = {
    'type': 'object',
    'properties': {
        'lang': {'type': 'string'},
        'source': {'type': 'string'},
        'all_fatal': {'type': 'boolean'},
        'compile': {
            'type': 'object',
            'properties': ISOLATE_OPTS_PROPERTIES,
        },
        'execute': {
            'type': 'object',
            'properties': EXECUTE_PROPERTIES,
        },
        'tests': {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    'name': {'type': 'string'},
                    'fatal': {'type': 'boolean'},
                    **EXECUTE_PROPERTIES,
                },
            },
        },
    },
    'required': ['lang', 'source'],
}


def validate(json):
    return jsonschema.validate(json, SCHEMA)



from camisole.models import Lang, Program


class VisualBasic(Lang):
    source_ext = '.vb'
    compiler = Program('vbnc', opts=['/optimize+'], version_opt='/help')
    interpreter = Program('mono')
    reference_source = r'''
Imports System
Public Module modmain
Sub Main()
    Console.WriteLine ("42")
End Sub
End Module
'''

    def compile_opt_out(self, output):
        return ['/out:' + output]



import camisole.languages


async def test(lang_name):
    expected = '42\n'
    lang_cls = camisole.languages.by_name(lang_name)
    lang = lang_cls({'source': lang_cls.reference_source, 'tests': [{}]})
    raw_result = await lang.run()
    try:
        stdout = raw_result['tests'][0]['stdout']
        if stdout == expected:
            return True, raw_result
        return False, raw_result
    except (KeyError, IndexError, ValueError):
        return False, raw_result



import math
import os
import textwrap


def uniquify(seq):
    seen = set()
    return (x for x in seq if not (x in seen or seen.add(x)))


def indent(text, n=4):
    return textwrap.indent(text, ' ' * n)


def parse_size(str_size):
    """
    Transforms "1K", "1k" or "1 kB"-like strings to actual integers.
    Returns None if input is None.
    """
    if str_size is None:
        return None
    str_size = str_size.lower().strip().rstrip('b')
    for l, m in (('k', 1 << 10), ('m', 1 << 20), ('g', 1 << 30)):
        if str_size.endswith(l):
            return int(str_size.rstrip(l)) * m
    return int(str_size)


def parse_float(str_float):
    if str_float is None:
        return None
    return float(str_float)


def tabulate(rows, headers=None, fmt="", margin=1):
    ncols = len(rows[0])
    lengths = [-math.inf] * ncols
    if headers:
        # don't side-effect modify rows
        rows = [headers] + rows
    for row in rows:
        lengths = [max(l, len(col)) for l, col in zip(lengths, row)]
    lengths = [l + margin for l in lengths]
    if not fmt:
        fmt = "".join("{:<{s%d}}%s" % (i, " | " if i < ncols  - 1 else "")
                      for i in range(ncols))
    for row in rows:
        yield fmt.format(*row, **{f's{i}': l for i, l in enumerate(lengths)})


def which(binary):
    search_prefixes = ['/usr', '/lib', '/bin']
    path = [*os.environ.get('PATH').split(os.pathsep),
            '/usr/bin',
            '/usr/local/bin'
            '/bin',
    ]
    if os.path.dirname(binary) and os.access(binary, os.X_OK):
        return binary
    for part in path:
        # Ignore matches that are not inside standard directories
        if not any(part.startswith(prefix) for prefix in search_prefixes):
            continue
        p = os.path.join(part, binary)
        if os.access(p, os.X_OK):
            return p
    return binary



from camisole.models import Lang, Program


class OCaml(Lang):
    source_ext = '.ml'
    compiler = Program('ocamlopt', opts=['-w', 'A'], version_opt='-v')
    reference_source = r'print_int 42; print_string "\n";'



# This file is part of Camisole.
#
# Copyright (c) 2016 Antoine Pietri <antoine.pietri@prologin.org>
# Copyright (c) 2016 Association Prologin <info@prologin.org>
#
# Camisole is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Prologin-SADM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Prologin-SADM.  If not, see <http://www.gnu.org/licenses/>.

import functools
import os
import subprocess
import sys

from camisole.utils import parse_size, parse_float


def lscpu():
    out = subprocess.check_output('lscpu')
    out = out.decode().strip().split('\n')
    return {k: v.strip() for line in out for k, v in (line.split(':', 1),)}


def meminfo():
    with open('/proc/meminfo') as f:
        out = f.read().strip().split('\n')
    return {k: v.strip() for line in out for k, v in (line.split(':', 1),)}


# This function only gives static system information so we can cache it
# indefinitely.
@functools.lru_cache()
def info():
    uname = os.uname()
    cpu = lscpu()
    mem = meminfo()
    return {
        'arch': uname.machine,
        'byte_order': sys.byteorder,
        'cpu_cache_L1d': parse_size(cpu.get('L1d cache')),
        'cpu_cache_L1i': parse_size(cpu.get('L1i cache')),
        'cpu_cache_L2': parse_size(cpu.get('L2 cache')),
        'cpu_cache_L3': parse_size(cpu.get('L3 cache')),
        'cpu_count': os.cpu_count(),
        'cpu_mhz': parse_float(cpu.get('CPU MHz')),
        'cpu_name': cpu.get('Model name'),
        'kernel': uname.sysname,
        'kernel_release': uname.release,
        'kernel_version': uname.version,
        'memory': parse_size(mem.get('MemTotal'))
    }



from camisole.languages.c import C
from camisole.models import Lang, PipelineLang, Program
from camisole.utils import which


class BrainfuckToC(Lang, register=False):
    source_ext = '.bf'
    extra_binaries = {'esotope': Program('esotope-bfc', version_opt=None)}
    compiler = Program('python2',
                       env={'PYTHONPATH': '/usr/lib/python2.7/site-packages'},
                       opts=['-S', extra_binaries['esotope'].cmd, '-v', '-fc'])

    def compile_opt_out(self, output):
        # there is no way to specify an output file, use stderr
        return []

    def read_compiled(self, path, isolator):
        return isolator.stdout.encode()


class Brainfuck(PipelineLang):
    source_ext = '.bf'
    sub_langs = [BrainfuckToC, C]
    reference_source = r'-[----->+<]>+.--.>++++++++++.'



from camisole.models import Lang, Program


class Rust(Lang):
    source_ext = '.rs'
    compiler = Program('rustc', opts=['-W', 'warnings', '-O'])
    reference_source = r'''
fn main() {
    println!("42");
}'''



from camisole.models import Lang, Program


class CXX(Lang, name="C++"):
    source_ext = '.cc'
    compiler = Program('g++', opts=['-std=c++14', '-Wall', '-Wextra', '-O2'])
    reference_source = r'''
#include <iostream>
int main()
{
    std::cout << 42 << std::endl;
    return 0;
}
'''



from camisole.models import Lang, Program


class Ada(Lang):
    source_ext = '.adb'
    compiler = Program('gnatmake', opts=['-f'])
    reference_source = r'''
with Ada.Text_IO; use Ada.Text_IO;
procedure Hello is
begin
   Put_Line("42");
end Hello;
'''



# This file is part of Camisole.
#
# Copyright (c) 2016 Antoine Pietri <antoine.pietri@prologin.org>
# Copyright (c) 2016 Association Prologin <info@prologin.org>
#
# Camisole is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Prologin-SADM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Prologin-SADM.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import itertools
import logging
import pathlib
import subprocess
import tempfile


async def communicate(cmdline, data=None, **kwargs):
    logging.debug('Running %s', ' '.join(cmdline))
    proc = await asyncio.create_subprocess_exec(
        *cmdline, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
        stderr=subprocess.PIPE, **kwargs)
    stdout, stderr = await proc.communicate(data)
    retcode = await proc.wait()
    return retcode, stdout, stderr


OPTIONS = [
    'fsize',
    'mem',
    'processes',
    'quota',
    'stack'
    'time',
    'wall-time',
]

# TODO(seirl): find a way to get those from the config?
NUM_BOXES = 1000
PATH_BOXES = pathlib.Path('/var/lib/isolate')

BOX_ID_LOCK = asyncio.Lock()


class Isolator:
    def __init__(self, opts, allowed_dirs=None):
        self.opts = opts
        self.allowed_dirs = allowed_dirs if allowed_dirs is not None else []
        self.path = None
        self.cmd_base = None

        # Directory containing all the info of the program
        self.stdout_file = '._stdout'
        self.stderr_file = '._stderr'
        self.meta_file = None

        self.stdout = None
        self.stderr = None
        self.meta = None
        self.info = None

        # Result of the isolate binary
        self.isolate_retcode = None
        self.isolate_stdout = None
        self.isolate_stderr = None

    async def __aenter__(self):
        with (await BOX_ID_LOCK):
            busy = {int(p.name) for p in PATH_BOXES.iterdir()}
            avail = set(range(NUM_BOXES)) - busy
            try:
                self.box_id = avail.pop()
            except KeyError:
                raise RuntimeError("No isolate box ID available.")
            self.cmd_base = ['isolate', '--box-id', str(self.box_id), '--cg']

            cmd_init = self.cmd_base + ['--init']
            retcode, stdout, stderr = await communicate(cmd_init)
            if retcode != 0:
                raise RuntimeError("{} returned code {}: “{}”".format(
                    cmd_init, retcode, stderr))
            self.path = pathlib.Path(stdout.strip().decode()) / 'box'
            self.meta_file = tempfile.NamedTemporaryFile(prefix='camisole-meta-')
            self.meta_file.__enter__()

    async def __aexit__(self, exc, value, tb):
        meta_defaults = {
            'cg-mem': 0,
            'csw-forced': 0,
            'csw-voluntary': 0,
            'exitcode': 0,
            'exitsig': None,
            'killed': False,
            'max-rss': 0,
            'message': None,
            'status': 'OK',
            'time': 0.0,
            'time-wall': 0.0,
        }
        with open(self.meta_file.name) as f:
            m = (l.strip() for l in f.readlines())
        m = dict(l.split(':', 1) for l in m if l)
        m = {k: (type(meta_defaults[k])(v)
                 if meta_defaults[k] is not None else v)
             for k, v in m.items()}
        self.meta = {**meta_defaults, **m}
        verbose_status = {
            'OK': 'OK',
            'RE': 'RUNTIME_ERROR',
            'TO': 'TIMED_OUT',
            'SG': 'SIGNALED',
            'XX': 'INTERNAL_ERROR',
        }
        self.meta['status'] = verbose_status[self.meta['status']]

        self.info = {
            'stdout': self.stdout,
            'stderr': self.stderr,
            'exitcode': self.isolate_retcode,
            'meta': self.meta
        }

        cmd_cleanup = self.cmd_base + ['--cleanup']
        retcode, stdout, stderr = await communicate(cmd_cleanup)
        if retcode != 0:
            raise RuntimeError("{} returned code {}: “{}”".format(
                cmd_cleanup, retcode, stderr))

        self.meta_file.__exit__(exc, value, tb)

    async def run(self, cmdline, data=None, env=None, **kwargs):
        cmd_run = self.cmd_base[:]
        cmd_run += list(itertools.chain(
            *[('-d', d) for d in self.allowed_dirs]))

        for opt in OPTIONS:
            v = self.opts.get(opt)
            if v is not None:
                cmd_run.append(f'--{opt}={v!s}')
            # Unlike isolate, we don't limit the number of processes by default
            elif opt == 'processes':
                cmd_run.append('-p')

        for e in ['PATH', 'LD_LIBRARY_PATH']:
            cmd_run += ['--env', e]

        for key, value in (env or {}).items():
            cmd_run += ['--env={}={}'.format(key, value)]

        cmd_run += [
            '--meta={}'.format(self.meta_file.name),
            '--stdout={}'.format(self.stdout_file),
            '--stderr={}'.format(self.stderr_file),
            '--run', '--'
        ]
        cmd_run += cmdline

        self.isolate_retcode, self.isolate_stdout, self.isolate_stderr = (
            await communicate(cmd_run, data=data, **kwargs))

        try:
            with (self.path / self.stdout_file).open(errors='ignore') as f:
                self.stdout = f.read()
            with (self.path / self.stderr_file).open(errors='ignore') as f:
                self.stderr = f.read()
        except (IOError, PermissionError):
            # Something went wrong, isolate was killed before changing the
            # permissions or unreadable stdout/stderr
            self.stdout = ''
            self.stderr = ''



import sys
from setuptools import setup, find_packages

if sys.version_info < (3, 6):
    sys.exit("Camisole requires Python 3.6 or later.")

setup(
    name='camisole',
    version='0.5',
    packages=find_packages(),
    url='https://github.com/prologin/camisole',
    license='GPL',
    author='Antoine Pietri, Alexandre Macabies',
    description='An asyncio-based source compiler and test runner.',
    install_requires=['aiohttp'],
    setup_requires=['pytest-runner'],
    tests_require=['pytest', 'pytest-cov'],
    test_suite='pytest',
    entry_points={
        'console_scripts': ['camisole = camisole.__main__:main'],
    },
)



import aiohttp.web
import functools
import json
import jsonschema
import traceback

import camisole.languages
import camisole.ref
import camisole.schema
import camisole.system


def json_handler(wrapped):
    @functools.wraps(wrapped)
    async def wrapper(request):
        # XXX: hack no longer needed in recent aiohttp versions
        if not hasattr(request, 'query'):
            import urllib.parse
            request.query = urllib.parse.parse_qs(request.query_string,
                                                  keep_blank_values=True)
            request.query = {k: v[-1] for k, v in request.query.items()}

        try:
            try:
                data = await request.text()
                data = json.loads(data) if data else {}
            except json.decoder.JSONDecodeError:
                raise RuntimeError('Invalid JSON')
            result = await wrapped(request, data)
        except Exception:
            result = {'success': False, 'error': traceback.format_exc()}
        else:
            result = {'success': True, **result}

        pretty = ('pretty' in request.query
                  and request.query['pretty'].lower() in ('', 'true', '1'))
        indent = 4 if pretty else None
        result = json.dumps(result, sort_keys=True, indent=indent)
        return aiohttp.web.Response(body=result.encode(),
                                    content_type='application/json')
    return wrapper


@json_handler
async def run_handler(request, data):
    try:
        camisole.schema.validate(data)
    except jsonschema.exceptions.ValidationError as e:
        return {'success': False, 'error': str(e)}

    lang_name = data['lang'].lower()
    try:
        lang = camisole.languages.by_name(lang_name)(data)
    except KeyError:
        raise RuntimeError('Incorrect language {}'.format(lang_name))

    return await lang.run()


@json_handler
async def test_handler(request, data):
    langs = camisole.languages.all().keys()
    langs -= set(data.get('exclude', []))

    results = {name: {'success': success, 'raw': raw}
               for name in langs
               for success, raw in [await camisole.ref.test(name)]}
    return {'results': results}


@json_handler
async def system_handler(request, data):
    return {'system': camisole.system.info()}


@json_handler
async def languages_handler(request, data):
    return {'languages': {lang: {'name': cls.name, 'programs': cls.programs()}
            for lang, cls in camisole.languages.all().items()}}

async def default_handler(request):
    return aiohttp.web.Response(
        text="Welcome to Camisole. Use the /run endpoint to run some code!\n")


def make_application(**kwargs):
    app = aiohttp.web.Application(**kwargs)
    app.router.add_route('POST', '/run', run_handler)
    app.router.add_route('*', '/', default_handler)
    app.router.add_route('*', '/languages', languages_handler)
    app.router.add_route('*', '/system', system_handler)
    app.router.add_route('*', '/test', test_handler)
    return app


def run(**kwargs):  # noqa
    app = make_application()
    aiohttp.web.run_app(app, **kwargs)



import re
import subprocess
from pathlib import Path

from camisole.models import Lang, Program

RE_WRONG_FILENAME_ERROR = re.compile(r'error:\s+class\s+(.+?)\s+is\s+public,')
PSVMAIN_SIGNATURE = 'public static void main('
PSVMAIN_DESCRIPTOR = 'descriptor: ([Ljava/lang/String;)V'


class Java(Lang):
    source_ext = '.java'
    compiled_ext = '.class'
    compiler = Program('javac', env={'LANG': 'C'}, version_opt='-version')
    interpreter = Program('java', version_opt='-version')
    # /usr/lib/jvm/java-8-openjdk/jre/lib/amd64/jvm.cfg links to
    # /etc/java-8-openjdk/amd64/jvm.cfg
    allowed_dirs = ['/etc/java-8-openjdk']
    # ensure we can parse the javac(1) stderr
    extra_binaries = {'disassembler': Program('javap', version_opt='-version')}
    reference_source = r'''
class SomeClass {
    static int fortytwo() {
        return 42;
    }
    static class Subclass {
        // nested psvmain! wow!
        public static void main(String args[]) {
           System.out.println(SomeClass.fortytwo());
        }
    }
}
'''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # use an illegal class name so that javac(1) will spit out the actual
        # class named used in the source
        self.class_name = '1337'
        # we give priority to the public class, if any, so keep a flag if we
        # found such a public class
        self.found_public = False

        # Don't even try to cgroup the java process:
        # http://stackoverflow.com/questions/19910468
        # https://bugs.launchpad.net/ubuntu/+source/openjdk-7/+bug/1241926
        # https://bugs.openjdk.java.net/browse/JDK-8071445
        # http://bugs.java.com/view_bug.do?bug_id=8043516

        # Instead, pass the memory limit as the maximum heap size of the
        # runtime.
        try:
            self.heapsize = self.opts['execute'].pop('mem')
        except KeyError:
            self.heapsize = None

    def compile_opt_out(self, output):
        # javac has no output directive, file name is class name
        return []

    async def compile(self):
        # try to compile with default class name (Main)
        retcode, info, binary = await super().compile()
        if retcode != 0:
            # error: public class name is not '1337' -- obviously, it's illegal,
            # so find what it actually is
            match = RE_WRONG_FILENAME_ERROR.search(info['stderr'])
            if match:
                self.found_public = True
                self.class_name = match.group(1)
                # retry with new name
                retcode, info, binary = await super().compile()
        return (retcode, info, binary)

    def source_filename(self):
        return self.class_name + self.source_ext

    def execute_filename(self):
        # return eg. Main.class
        return self.class_name + self.compiled_ext

    def execute_command(self, output):
        cmd = [self.interpreter.cmd]

        # Use the memory limit as a maximum heap size
        if self.heapsize is not None:
            cmd.append(f'-Xmx{self.heapsize}k')

        # foo/Bar.class is run with $ java -cp foo Bar
        cmd += ['-cp', str(Path(self.filter_box_prefix(output)).parent),
                self.class_name]
        return cmd

    def find_class_having_main(self, classes):
        for file in classes:
            # run javap(1) with type signatures
            try:
                stdout = subprocess.check_output(
                    [self.extra_binaries['disassembler'].cmd, '-s', str(file)],
                    stderr=subprocess.DEVNULL, env=self.compiler.env)
            except subprocess.SubprocessError:
                continue
            # iterate on lines to find p s v main() signature and then
            # its descriptor on the line below; we don't rely on the type
            # from the signature, because it could be String[], String... or
            # some other syntax I'm not even aware of
            lines = iter(stdout.decode().split('\n'))
            for line in lines:
                if line.lstrip().startswith(PSVMAIN_SIGNATURE):
                    if next(lines).lstrip() == PSVMAIN_DESCRIPTOR:
                        return file.stem

    def read_compiled(self, path, isolator):
        # in case of multiple or nested classes, multiple .class files are
        # generated by javac
        classes = list(isolator.path.glob('*.class'))
        files = [(file.name, file.open('rb').read()) for file in classes]
        if not self.found_public:
            # the main() may be anywhere, so run javap(1) on all .class
            new_class_name = self.find_class_having_main(classes)
            if new_class_name:
                self.class_name = new_class_name
        return files

    def write_binary(self, path, binary):
        # see read_compiled(), we need to write back all .class files
        # but give only the main class name (execute_filename()) to java(1)
        for file, data in binary:
            with (path / file).open('wb') as c:
                c.write(data)
        return path / self.execute_filename()
